// -*- coding:euc-jp -*-
#define _GNU_SOURCE
#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iconv.h>
#include <stdbool.h>
#include <libgen.h>
#include <sys/stat.h>
#include <errno.h>
#include "ut.h"
#include <ctype.h>


static char* cur_to_u8(char* out,const char* in,int inlen);
static char* no_conv(char* out,const char* in,int inlen);


//e-a2a2 s-81a0 u-25a1 t-e296a1 ''
#define TOFU_E	0xa2a2
#define TOFU_S	0xa081
#define TOFU_U2	0xa125
#define TOFU_U8	0xa196e2

//(from,to) primary,secondary
enum{ EU16_08,EU16_13,
      U16E_08,U16E_13,
      EU8_08,EU8_13,
      U16S,U16S_13,
      SU16,SU16_13,
      U8E_08,U8E_13,
      U16U8,CV_DUM0,
      U8U16,CV_DUM1,
      CNVMAX};
iconv_t Icv[CNVMAX];

extern iconv_t CurrentToUtf8;

__attribute__((constructor))
void iconv_ini(void)
{
    const char UTF16[] = "utf-16le",UTF8[]="utf-8",
	EUCJP[]="euc-jp-ms",EUCX0213[]="euc-jisx0213",
	SJIS[]="shift_jis",SJIS13[]="shift_jisx0213";
    const char *cs[][2]={ /* to,from */
	{UTF16,EUCJP},		{UTF16,EUCX0213},
	{EUCJP,UTF16},		{EUCX0213,UTF16},
	{UTF8,EUCJP},		{UTF8,EUCX0213},
	{SJIS,UTF16},		{SJIS13,UTF16},
	{UTF16,SJIS},		{UTF16,SJIS13},
	{EUCJP,UTF8},		{EUCX0213,UTF8},
	{UTF8,UTF16},		{NULL,NULL},
	{UTF16,UTF8},		{NULL,NULL},
    };
    for(unsigned n=0; n<ITEMS(Icv); ++n){
	if(cs[n][0]!=NULL){
	    Icv[n] = iconv_open(cs[n][0],cs[n][1]);
#ifdef __FreeBSD__
	    if(Icv[n] == (iconv_t)-1){
		for(int ft=0; ft<2; ++ft){
		    if(cs[n][ft]==EUCX0213)
			cs[n][ft]=EUCJP;
		    if(cs[n][ft]==SJIS13)
			cs[n][ft]=SJIS;
		}
		Icv[n] = iconv_open(cs[n][0],cs[n][1]);
	    }
#endif
	}
    }

    //ȥ뢪utf8
    char* loc = getenv("LC_CTYPE");
    if(loc == NULL)
	loc = getenv("LANG");
    if(loc == NULL || (loc=strchr(loc,'.')) == NULL)
	loc = strdup("utf8");
    else{
	loc = strdup(loc+1);
	for(char* lp; (lp=strchr(loc,'-'))!=NULL; StrDel(lp,0,1))
	    ;
    }
    if(strcasecmp(loc,"utf8") == 0)
	CurToU8 = no_conv;
    else if(strcasecmp(loc,"eucjp") == 0)
	CurToU8 = EjToU8;
    else if(strcasecmp(loc,"sjis") == 0)
	CurToU8 = SjToU8;
    else{
	CurrentToUtf8 = iconv_open("utf8",loc);
	CurToU8 = CurrentToUtf8!=(iconv_t)-1 ? cur_to_u8 : no_conv;
    }
    free(loc);
}

iconv_t CurrentToUtf8;
char* (*CurToU8)(char* out,const char* in,int inlen);
//ȥ뢪utf8
static char* cur_to_u8(char* out,const char* in,int inlen)
{
    size_t ileft = inlen<0 ? strlen(in) : inlen;
    char* ind = strndup(in,ileft);
    char* ind0 = ind;
    size_t oleft = ileft*3+1;
    if(out == NULL)
	out = malloc(oleft);
    char* out0 = out;
    iconv(CurrentToUtf8,&ind,&ileft,&out,&oleft);
    *out = 0;
    free(ind0);
    return out0;
}
static char* no_conv(char* out,const char* in,int inlen)
{
    if(inlen < 0)
	inlen = strlen(in);
    if(out == NULL)
	out = strndup(in,inlen);
    else
	*(strncpy(out,in,inlen)+inlen) = 0;
    return out;
}

__attribute__((destructor))
void iconv_fin(void)
{
    for(unsigned n=0; n<ITEMS(Icv); ++n)
	if(Icv[n] != NULL){
	    iconv_close(Icv[n]);
	    Icv[n] = NULL;
	}
}


uint16_t* Swap2p(void* x,int dist)
{
    unsigned char *a,*b,c;
    a = (unsigned char*)x;
    b = a+dist;
    c = *a; *a = *b; *b = c;
    return (uint16_t*)x;
}

uint16_t Swap2(uint16_t x)
{
    return *Swap2p(&x,1);
}

uint16_t Swap2c(const void* x)
{
    return Swap2(*(const uint16_t*)x);
}

unsigned char* rev_copy(unsigned char* to,const unsigned char* from,int len)
{
    unsigned char* r = to;
    from += len;
    while(--len >= 0)
	*(to++) = *(--from);
    return r;
}

int* RevInt(int* x)
{
    int x0 = *x;
    return (int*)rev_copy((unsigned char*)x,(const unsigned char*)&x0,sizeof(int));
}

/*
  wchar_t礭gccwinegccǰ㤦
  ̤Υǥ쥯ȥgccȤǽ⤢Τǡ𤷤ʤ褦wchar_t,wcs...()
  Ȥ鷺˼δؿ뤳Ȥˤ롣
*/
size_t WcLen(const uint16_t* s)
{
    const uint16_t* s0 = s;

    if(s == NULL)
	return 0;

    while(*(s++) != 0)
	;
    return s-s0-1;
}

uint16_t* WcChr(uint16_t* s,uint16_t c)
{
    while(*s!=0 && *s!=c)
	++s;
    return (*s!=0 || c==0) ? s : NULL;
}

uint16_t* WcCpy(uint16_t* dst,const uint16_t* src)
{
    return memcpy(dst,src,(WcLen(src)+1)*2);
}

//ڡڤΤߤб
uint16_t* U16Tok(uint16_t** src)
{
    const int spc = ' '; //wejΤȤ(' '<<8)
    while(**src == spc)
	++ *src;
    if(**src == 0)
	return NULL;
    uint16_t* start = *src;
    while(**src!=0 && **src!=spc)
	++ *src;
    if(**src != 0)
	*((*src)++) = 0;
    return start;
}
    
/*
  weucjpwcharʸcharʸˤ롣
  ͤfree뤳ȡ
  cannaʤΤsrcβӥåȤ裱Хȡ壸ӥåȤ裲ХȤˤʤ롣
  Ⱦѥʤ0x8eĤƤʤ(裱ХȤϣ)
  cannajisx0212򣲥ХȤǰ裱ХȤѴ褦
*/
char* ToMb(const uint16_t* src)
{
    Array dst;
    
    ArNew(&dst,1,NULL);
    if(src != NULL){
	for(; *src!=0; ++src){
	    unsigned char *dp;
	    if((*src & 0xff) != 0){
		if((*src>>8) & 0x80){
		    *(uint16_t*)ArExpand(&dst,2) = *src;
		}else{ //jisx0212
		    *(dp = ArExpand(&dst,3)) = 0x8f;
		    *(uint16_t*)(dp+1) = *src|0x8000;
		}
	    }else{
		unsigned char uc = (unsigned char)(*src>>8);
		if(uc>=0xa1 && uc<=0xdf){ //Ⱦѥ
		    dp = ArExpand(&dst,2);
		    *(dp++) = 0x8e;
		}else
		    dp = ArExpand(&dst,1);
		*dp = uc;
	    }
	}
	*(char*)ArExpand(&dst,1) = 0;
    }
    return ArAdr(&dst);
}

/*
  eucjpcanna wcharˤ롣
  cannaʤΤsrcβӥåȤ裱Хȡ壸ӥåȤ裲ХȤˤʤ롣
  Ⱦѥʤ0x8eĤƤʤ(裱ХȤϣ) [r140 ??? 0x8eʤʸ롣]
  cannajisx0212򣲥ХȤǰ裱ХȤѴ褦
  src_len<0ΤȤdst_len˥̥ʸޤࡣsrc_lenꤷȤϤϰϤǽʸˤʤ롣
  src==NULLΤȤϥ顼ȤNULL֤
*/
uint16_t* ToWc(uint16_t* dst,int* dst_len,const char* src,int src_len)
{
    uint16_t *dst0,wc;

    if(src == NULL)
	return NULL;
    if(dst == NULL)
	dst = calloc(src_len<0 ? EjLen(src)+1:src_len,2);
    dst0 = dst;

    while(src_len>0 || (src_len<0 && *src!=0)){
	switch((uint8_t)*src){
	case 0x8e: //Ⱦѥ
	    wc = (((uint16_t)*(++src))<<8)|0x8e; //??? 0x8eʤʸ褦[r140]
	    ++src;
	    src_len -= 3;
	    break;
	case 0x8f: //0212
	    wc = *(uint16_t*)(++src) & 0x7fff;
	    src += 2;
	    src_len -= 2;
	    break;
	case 0xa1 ... 0xff: //0208
	    wc = *(uint16_t*)src;
	    src += 2;
	    src_len -= 2;
	    break;
	default: //ascii
	    wc = ((uint16_t)*(src++))<<8;
	    -- src_len;
	}
	*(dst++) = wc;
    }
    if(src_len < 0) //̥ʸޤǤλϽϤ˥̥դ롣
	*(dst++) = 0;
    if(dst_len != NULL)
	*dst_len = dst-dst0;
    return dst0;
}

uint16_t* WcDup(const uint16_t* src)
{
    return WcCpy(calloc(WcLen(src)+1,2),src);
}

/*
  utf16ʸ޻ˤ롣
  ѴǤʤäȤNULL֤롣
*/
const char* U16Zen2Romaji(uint16_t u16)
{
    static const char* hira[]={//0x3041
	"xa"/**/,	"a"/**/,	"xi"/**/,	"i"/**/,	"xu"/**/,	"u"/**/,
 	"xe"/**/,	"e"/**/,	"xo"/**/,	"o"/**/,	"ka"/**/,	"ga"/**/,
 	"ki"/**/,	"gi"/**/,	"ku"/**/,	"gu"/**/,	"ke"/**/,	"ge"/**/,
 	"ko"/**/,	"go"/**/,	"sa"/**/,	"za"/**/,	"si"/**/,	"ji"/**/,
 	"su"/**/,	"zu"/**/,	"se"/**/,	"ze"/**/,	"so"/**/,	"zo"/**/,
 	"ta"/**/,	"da"/**/,	"ti"/**/,	"di"/**/,	"xtu"/**/,	"tu"/**/,
 	"du"/**/,	"te"/**/,	"de"/**/,	"to"/**/,	"do"/**/,	"na"/**/,
 	"ni"/**/,	"nu"/**/,	"ne"/**/,	"no"/**/,	"ha"/**/,	"ba"/**/,
 	"pa"/**/,	"hi"/**/,	"bi"/**/,	"pi"/**/,	"fu"/**/,	"bu"/**/,
 	"pu"/**/,	"he"/**/,	"be"/**/,	"pe"/**/,	"ho"/**/,	"bo"/**/,
 	"po"/**/,	"ma"/**/,	"mi"/**/,	"mu"/**/,	"me"/**/,	"mo"/**/,
 	"xya"/**/,	"ya"/**/,	"xyu"/**/,	"yu"/**/,	"xyo"/**/,	"yo"/**/,
 	"ra"/**/,	"ri"/**/,	"ru"/**/,	"re"/**/,	"ro"/**/,	"xwa"/**/,
 	"wa"/**/,	"wi"/**/,	"we"/**/,	"wo"/**/,	"nn"/**/,
	"vu"/**/, 	"xka"/*()*/,	"xke"/*()*/ //0x3096
    };
    static char ascii_buf[2]={0,0};

    const char* rmj;
    switch(u16){
    case 0 ... 0x7f: //ascii
	ascii_buf[0] = u16;
	rmj = ascii_buf;
	break;
    case 0x2018: //󥰥륯ơ
	rmj = "`";
	break;
    case 0x2019: //󥰥륯ơ
	rmj = "'";
	break;
    case 0x201d: //֥륯ơ
	rmj = "\"";
	break;
    case 0x3000: //ѥڡ
	rmj = " ";
	break;
    case 0x3001: //
	rmj = ",";
	break;
    case 0x3002: //
	rmj = ".";
	break;
    case 0x300c: //
	rmj = "[";
	break;
    case 0x300d: //
	rmj = "]";
	break;
    case 0x30a1 ... 0x30f6: //ʡҤ餬ʤν³
	u16 -= 0x60;
    case 0x3041 ... 0x3096: //Ҥ餬
	rmj = hira[u16-0x3041];
	break;
    case 0x30fc: //Katakana-Hiragana Prolonged Sound Mark
	rmj = "-";
	break;
    case 0xff01 ... 0xff5e: //Ѥε桢ե٥å
	ascii_buf[0] = u16-0xff01+0x21;
	rmj = ascii_buf;
	break;
    case 0xffe3:
	rmj = "~";
	break;
    default:
	rmj = NULL;
    }
    return rmj;
}

#if 0
/*
  eucjpȾѥ --> ѤҤ餬
  syn=true
  zen_asc=asciiѤˤȤtrue
  ͡dst
	use_src=ȤäХȿιǣʸȤȤ⤢
  dstˤϺ磳Хɬ
*//*
char* EjHan2Zen(char* dst,const char* src0,int* use_src,bool syn,bool zen_asc)
{
    static const char as20[]=
	"ɡ"	"ʡˡܡݡ"	//20
	""	"䡩"	//30
	"£ãģţƣ"	"ȣɣʣˣ̣ͣΣ"	//40
	"Уѣңӣԣգ֣"	"أ٣ڡΡϡ"	//50
	"ƣ"	""	//60
	""	"Сá\xa1\xc1"	//70
	;
    static const char wk0[]=
	"֡ס"		""	//a1
	""	""	//b0
	"ĤƤȤʤˤ"	"ͤΤϤҤդؤۤ"	//c0
	"ߤ"	"󡫡"	//d0
	;
    static const char wd0[]=
	""		""	//a1
	""	""	//b0
	"¤ŤǤɢ"	"ФӤ֤٤ܢ"	//c0
	""	""	//d0
	;
    static const char wp0[]={
	""		""	//a1
	""	""	//b0
	""	"ѤԤפڤݢ"	//c0
	""	""	//d0
    };
    static const uint16_t *wk=(const uint16_t*)wk0;
    static const uint16_t *wdp[]={(const uint16_t*)wd0,(const uint16_t*)wp0};

    const unsigned char *src = (const unsigned char*)src0;
    int ofs;
    uint16_t z,tofu='';
    unsigned char dp;

    switch(*src){
    case 0 ... 0x7f:
	if(zen_asc){
	    *(uint16_t*)dst = *(uint16_t*)(as20 + (*(src++)-0x20)*2);
	    dst += 2;
	}else
	    *(dst++) = *(src++);
	break;
    case 0x8e:  //ȾѥʤΥ
	++src;
	ofs = *(src++)-0xa1;
	if(syn && src[0]==0x8e && (dp=src[1]-0xde,dp<2) && (z=wdp[dp][ofs])!=tofu){
	    *(uint16_t*)dst = z;
	    src += 2;
	}else{
	    *(uint16_t*)dst = wk[ofs];
	}
	dst += 2;
	break;
    case 0x8f:
	*(dst++) = *(src++);
    default:
	*(uint16_t*)dst = *(uint16_t*)src;
	dst += 2;
	src += 2;
    }
    *use_src = src - (const unsigned char*)src0;
    return dst;
}
  */
/*
  eucjpȾѥ --> ѤҤ餬
  srclen=Хȿ
  syn=true
  zen_asc=asciiѤˤȤtrue
  asciiʸˤ뤿dstsrcĹʤǽ롣
*//*
char* HanToZen(char* dst,const char* src,int srclen,bool syn,bool zen_asc)
{
    char *dst0;
    int use_src;

    if(dst == NULL)
	dst = malloc((srclen>=0 ? srclen : EjLen(src)*3)+1);
    dst0 = dst;

    if(srclen < 0)
	srclen = INT_MAX;

    while(*src!=0 && srclen>0){
	dst = EjHan2Zen(dst,src,&use_src,syn,zen_asc);
	srclen -= use_src;
	src += use_src;
    }
    *dst = 0;
    return dst0;
}
*/
#endif

//eucjpʸƬnʸư롣
char* ForwardEj(char* ej,int n)
{
    while(--n >= 0 && *ej!=0){
	if(*ej == (char)0x8f)
	    ++ej;
	if(*(++ej) < 0)
	    ++ej;
    }
    return ej;
}

/*
  euc-jpʸ
*/
int EjLen(const char* ej)
{
    int len=0;
    if(ej != NULL){
	while(*ej != 0){
	    ej = ForwardEj((char*)ej,1);
	    ++len;
	}
    }
    return len;
}

/*
  numˤEU16_08,U16E_08Ϥȡ
  08ǥ顼Ф13ȤȤꤢʸɤ¸ϤǤ롣
  08˻ȤΤ''ԣʸѴǤʤݤϽФʤϤåƤ
  ??? iconv*in˽񤭹ꤹ
*/
bool conv(int num,char** in,size_t* ileft,char** out,size_t* oleft)
{
    bool st;
    char *in0 = *in, *out0=*out;
    size_t ileft0 = *ileft;

    st = (iconv(Icv[num],in,ileft,out,oleft)!=(size_t)-1);
    if(st && *out==out0){
	*in = in0;
	*ileft = ileft0;
	st = false;
    }
    if(!st && Icv[num+1]!=NULL){
	st = (iconv(Icv[num+1],in,ileft,out,oleft)!=(size_t)-1);
    }
    return st;
}

//euc-jp --> utf16
//dst0NULLΤȤmallocǳݤ롣
uint16_t* EjToU16(uint16_t* dst0,const char* src0)
{
    char *src,*src_orig;
    size_t ileft=strlen(src0),oleft=EjLen(src0)*2;
    uint16_t *dst;

    src = src_orig = strdup(src0);
    if(dst0 == NULL)
	dst0 = malloc(oleft+2);
    dst = dst0;
    while(!conv(EU16_08,&src,&ileft,(char**)&dst,&oleft)){
	*(dst++) = TOFU_U2;
	oleft -= 2;
	if(*src == (char)0x8f){
	    ++src;
	    --ileft;
	}
	src += 2;
	ileft -= 2;
    }
    *dst = 0;
    //iconv(Icv[E2U08],NULL,NULL,NULL,NULL);
    //iconv(Icv[E2U13],NULL,NULL,NULL,NULL);
    free(src_orig);
    return dst0;
}

//weucjp --> utf16
//dst0NULLΤȤmallocǳݤ롣
uint16_t* WejToU16(uint16_t* dst,const uint16_t* src)
{
    char* ej = ToMb(src);
    dst = EjToU16(dst,ej);
    free(ej);
    return dst;
}

/*
  utf16 --> ej or sj
  dst_lenϽϤʸsrc_len<0ΤȤϥ̥ʸʬޤࡣ
  src==NULLΤȤϥ顼ȤNULL֤
*/
char* u16_to_mb(int cv,int tofu,char* dst0,int* dst_len,const uint16_t* src0,int src_len)
{
    uint16_t *src,*src_orig;
    char *dst;
    size_t ileft,oleft;

    if(src0 == NULL)
	return NULL;
    if(src_len < 0)
	src_len = WcLen(src0)+1; //̥ʸޤ롣
    ileft = src_len*2;
    oleft = src_len*3; //ejѤ¿˳ݤ
    src = src_orig = memcpy(malloc(ileft+2),src0,ileft+2);
    if(dst0 == NULL)
	dst0 = malloc(oleft+1);
    dst = dst0;
    while(!conv(cv,(char**)&src,&ileft,&dst,&oleft)){
	*(uint16_t*)dst = tofu;
	dst += 2;
	oleft -= 2;
	++src;
	ileft -= 2;
    }
    if(dst_len != NULL)
	*dst_len = dst-dst0;
    free(src_orig);
    return dst0;
}

/* utf16 -->euc-jp
   ??? euc-jisx0213''ԣʸλϤ줺in-ptr,ileft롣
   euc-jpΤȤʤ櫓狼ʤ
   iconv(1)ǤϤɤäƤ
   shift-jisx0213ǤƱͤΤȤ롣u16'Ϥ'sj13Ѵin-ptr,ileft
   ʸʬ줿out-ptr,oleftϣʸʬʤä
*/
char* U16ToEj(char* dst,int* dst_len,const uint16_t* src,int src_len)
{
    return u16_to_mb(U16E_08,TOFU_E,dst,dst_len,src,src_len);
}

//utf16 --> weucjp
uint16_t* U16ToWej(uint16_t* dst,int* dst_len,const uint16_t* src,int src_len)
{
    int ejlen;
    
    char* ej = U16ToEj(NULL,&ejlen,src,src_len);
    dst = ToWc(dst,dst_len,ej,ejlen);
    free(ej);
    return dst;
}


//euc-jp --> utf8
char* EjToU8(char* dst,const char* src00,int ilen)
{
    if(src00 == NULL)
	return NULL;

    size_t ileft = ilen<0 ? strlen(src00) : ilen;
    size_t oleft=EjLen(src00)*8;
    char *dst0,*src0,*src;

    src0 = src = strdup(src00);
    if(dst == NULL)
	dst = malloc(oleft+1);
    dst0 = dst;
    while(!conv(EU8_08,&src,&ileft,&dst,&oleft)){
	*(int32_t*)dst = TOFU_U8;
	dst += 3;
	oleft -= 3;
	if(*src == (char)0x8f){
	    ++src;
	    --ileft;
	}
	src += 2;
	ileft -= 2;
    }
    *dst = 0;
    free(src0);
    return dst0;
}

//utf16 --> utf8
//src_len<0ΤȤdst_lenˤϥ̥ʸޤࡣ
char* U16ToU8(char* dst0,int* dst_len,const uint16_t* src0,int src_len)
{
    uint16_t *src,*src_orig;

    if(src0 == NULL)
	return NULL;
    if(src_len < 0)
	src_len = WcLen(src0)+1; //̥ʸޤ롣
    size_t ileft = src_len*2;
    size_t oleft = src_len*5;
    src = src_orig = memcpy(malloc(ileft+2),src0,ileft+2);
    if(dst0 == NULL)
	dst0 = malloc(oleft+1);
    char* dst = dst0;
    conv(U16U8,(char**)&src,&ileft,&dst,&oleft);
    if(dst_len != NULL)
	*dst_len = dst-dst0;
    free(src_orig);
    return dst0;
}

#if 0
//utf16 --> shift-jisx0213
/*
char* U16ToSj(char* out0,const uint16_t* in0,int in_len)
{
    return u16_to_mb(U16S,TOFU_S,out0,NULL,in0,in_len);
}

//ej --> sj
char* EjToSj(char* out,const char* in)
{
    uint16_t* u = EjToU16(NULL,in);
    out = U16ToSj(out,u,-1);
    free(u);
    return out;
}
*/
#endif

//!!! EjToU16Ȱˤ뤫
//sjis --> utf16
uint16_t* SjToU16(uint16_t* out0,const char* in0,size_t ileft)
{
    char *in,*in_orig;
    uint16_t *out;
    size_t oleft;

    in = in_orig = strdup(in0);
    if(ileft == (size_t)-1)
	ileft = strlen(in);
    oleft=ileft*2; //inȾѥʤξoutϣܤˤʤ롣ʸ롩

    if(out0 == NULL)
	out0 = malloc(oleft+2);
    out = out0;

    while(!conv(SU16,&in,&ileft,(char**)&out,&oleft)){
	*(out++) = TOFU_U2;
	oleft -= 2;
	in += 2;
	ileft -= 2;
    }
    *out = 0;
    free(in_orig);
    return out0;
}

//sjis --> ej
char* SjToEj(char* out,const char* in,int in_len)
{
    uint16_t* u = SjToU16(NULL,in,in_len);
    out = U16ToEj(out,NULL,u,-1);
    free(u);
    return out;
}

//sjis --> utf8
char* SjToU8(char* out,const char* in,int in_len)
{
    uint16_t* u = SjToU16(NULL,in,in_len);
    out = U16ToU8(out,NULL,u,-1);
    free(u);
    return out;
}

#if 0
/*
//sjis --> wej
uint16_t* SjToWej(uint16_t* out,const char* in,int in_len)
{
    char* ej = SjToEj(NULL,in,in_len);
    out = ToWc(out,NULL,ej,-1);
    free(ej);
    return out;
}
*/

/*
  eucjpѤҤ餬ʤȾѥʤˤ
  ͡Ϥʸ(1 or 2)
  dst˥̥ʸĤΤ,5ХɬסnullΤȤϽϤʤ
*//*
int EjZen2Han(char* dst,const char* src)
{
    //ѤҤ餬(''(0xa4a1)...''(0xa4f4))Ⱦѥ
    //1 1
    //2 Ⱦ
    //3 ȾѤΥ
    static uint16_t hira2hkana[]={
	'','','','','','','','','','','',1,'',1,'',1,'',1,'',1,
	'',1,'',1,'',1,'',1,'',1,'',1,'',1,'','',1,'',1,'',1,'',
	'','','','','',1,2,'',1,2,'',1,2,'',1,2,'',1,2,'','','',
	'','','','','','','','','','','','','','','','','',
	'','',3
    };

    union {
	uint8_t u1[4];
	uint16_t u2[2];
	uint32_t u4;
    } wc;
    int len=1,moji=1,idx;
    char dum_dst[5];

    if(dst == NULL)
	dst = dum_dst;
    
    wc.u1[1] = *(src++);
    if((wc.u1[1] & 0x80) == 0){
	//ascii
	*(dst++) = wc.u1[1];
	*dst = 0;
	return 1;
    }

    wc.u1[0] = *(src++);
    switch(wc.u2[0]){
    case '': wc.u1[1]=' '; break;
    case '': wc.u1[1]='!'; break;
    case '': wc.u1[1]='"'; break;
    case '': wc.u1[1]='#'; break;
    case '': wc.u1[1]='$'; break;
    case '': wc.u1[1]='%'; break;
    case '': wc.u1[1]='&'; break;
    case '': wc.u1[1]='\''; break;
    case '': wc.u1[1]='('; break;
    case '': wc.u1[1]=')'; break;
    case '': wc.u1[1]='*'; break;
    case '': wc.u1[1]='+'; break;
    case '': wc.u1[1]=','; break;
    case '': wc.u1[1]='-'; break;
    case '': wc.u1[1]='.'; break;
    case '': wc.u1[1]='/'; break;
    case '': wc.u1[1]=':'; break;
    case '': wc.u1[1]=';'; break;
    case '': wc.u1[1]='<'; break;
    case '': wc.u1[1]='='; break;
    case '': wc.u1[1]='>'; break;
    case '': wc.u1[1]='?'; break;
    case '': wc.u1[1]='@'; break;
    case '': wc.u1[1]='['; break;
    case '': wc.u1[1]='\\'; break;
    case '': wc.u1[1]=']'; break;
    case '': wc.u1[1]='^'; break;
    case '': wc.u1[1]='_'; break;
    case '': wc.u1[1]='`'; break;
    case '': wc.u1[1]='{'; break;
    case '': wc.u1[1]='|'; break;
    case '': wc.u1[1]='}'; break;
    case '': wc.u1[1]='~'; break;

    case '': wc.u2[0]=''; ++len; break;
    case '': wc.u2[0]=''; ++len; break;
    case '': wc.u2[0]=''; ++len; break;
    case '': wc.u2[0]=''; ++len; break;
    case '': wc.u2[0]=''; ++len; break;
    case '': wc.u2[0]=''; ++len; break;
    case '': wc.u2[0]=''; ++len; break;
    case '': wc.u2[0]=''; ++len; break;
    case '' ... '':
	wc.u1[1]=(uint8_t)(wc.u2[0]-''+'0'); break;
    case '' ... '':
	++len;
	wc.u2[0] = hira2hkana[idx=wc.u2[0]-''];
        switch(wc.u2[0]){
	case 1:
	    wc.u2[0] = hira2hkana[idx-1];
	    wc.u2[1] = '';
	    len += 2;
	    ++moji;
	    break;
	case 2:
	    wc.u2[0] = hira2hkana[idx-2];
	    wc.u2[1] = '';
	    len += 2;
	    ++moji;
	    break;
	case 3:
	    wc.u2[0] = '';
	    wc.u2[1] = '';
	    len += 2;
	    ++moji;
	}
	break;
    default:
	fprintf(stderr,"%s:illegal code 0x%04x\n",__FUNCTION__,wc.u2[0]);
    }
    for(idx=0; idx<4; idx+=2){
	dst[idx] = wc.u1[idx+1];
	dst[idx+1] = wc.u1[idx];
    }
    dst[len] = 0;
    return moji;
}
  */
/*
  eucjpѤҤ餬 --> Ⱦѥ
*//*
char* ZenToHan(char* dst,const char* src)
{
    char *dst0;

    if(dst == NULL)
	dst = malloc(strlen(src)*4+1);
    dst0 = dst;

    while(*src != 0){
	dst += EjZen2Han(dst,src)*2;
	src += 2;
    }
    *dst = 0;
    return dst0;
}
  */

/*
  eucjpѤҤ餬 --> ѥ
  src_len=ʸ
*//*
char* HiraToKata(char* dst,const char* src,int src_len)
{
    char *dst0;

    if(src_len < 0)
	src_len = INT_MAX;

    if(dst == NULL)
	dst = malloc(src_len<INT_MAX ? src_len*3+1 : strlen(src)+1);
    dst0 = dst;

    while(*src!=0 && src_len>0){
	switch((uint8_t)*src){
	case 0 ... 0x7f:
	    *(dst++) = *(src++);
	    break;
	case 0x8f:
	    memcpy(dst,src,3);
	    src += 3;
	    dst += 3;
	    break;
	case 0xa4: //Ҥ餬
	    *(dst++) = 0xa5;
	    ++src;
	    *(dst++) = *(src++);
	    break;
	default:
	    *(int16_t*)dst = *(int16_t*)src;
	    src += 2;
	    dst += 2;
	}
	--src_len;
    }
    *dst = 0;
    return dst0;
}
  */
#endif

int MkDir(const char* p0)
{
    if(p0[0]=='/' && p0[1]==0){
	return 1;
    }

    int r=0;
    char* p = strdup(p0);
    char* pp = strdup(p0);
    if(MkDir(dirname(pp))){
	r = (mkdir(p,0777)==0);
	if(r)
	    chmod(p,0777);
	else
	    if(errno==EEXIST)
		r=1;
    }
    free(p);
    free(pp);
    return r;
}

/*
  ե饰ʸ󲽤롣
  bitsκǸǤdescNULLˤ뤳ȡ
  bufνϤʤǥ1ˤ뤳ȡ bufNULLΤȤmallocǳݤ롣
 */
Array* FlagStr(unsigned flags,const BitDesc* bits,Array* buf)
{
    if(buf == NULL)
	buf = ArNew(NULL,1,NULL);
    if(flags == 0){
	ArPrint(buf,"(none)");
    }else{
	const char* sep="";
	for(unsigned n=0; bits[n].desc!=NULL; ++n){
	    if((flags & bits[n].mask)){
		ArPrint(buf,"%s%s",sep,bits[n].desc);
		flags &= ~bits[n].mask;
		sep="|";
	    }
	}
	if(flags!=0){
	    ArPrint(buf,"%s0x%x",sep,flags);
	}
    }
    return buf;
}

/*
  u16Ⱦѥʸǡsrc[1]Ⱦäsrc[0]ȹǤȤ()
  ʤ1Ⱦʤ2֤ʳλ0֤
  Ⱦѥ/asciiʳ̥λ-1֤
  u16ʤΤ''롣ejѴƦʸˤʤΤդ뤳ȡ
  kata_hira:Ҥ餬ʤˤ1,ʤˤ0
  combinefalseΤȤǤʤȤ롣ĤޤʸȽΤߤˤʤ롣
*/
int U16CombineHan(const uint16_t* src,int kata_hira,bool combine)
{
    #define D 1
    #define H 2
    const char hira_table[]={ //u-ff61  'Ǥ''ϣ
	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,
	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,
 	0/**/,	0/**/,	D/**/,	0/**/,	0/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,
 	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	0/**/,	0/**/,
 	0/**/,	0/**/,	0/**/,	D|H/**/,	D|H/**/,	D|H/**/,	D|H/**/,
 	D|H/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,
 	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/ /*u-ff9f*/
    };
    const char kata_table[]={ //u-ff61  'Ǥ''D
	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,
	D/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,
 	0/**/,	0/**/,	D/**/,	0/**/,	0/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,
 	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	D/**/,	0/**/,	0/**/,
 	0/**/,	0/**/,	0/**/,	D|H/**/,	D|H/**/,	D|H/**/,	D|H/**/,
 	D|H/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,	0/**/,
 	0/**/,	0/**/,	0/**/,	0/**/,	D/**/,	0/**/,	0/**/,	0/**/ /*u-ff9f*/
    };
    const char* voiced_table[]={kata_table,hira_table};
    
    if(src[0]>0 && src[0]<0x80) //ascii
	return 0;
    if(src[0]<0xff61 || src[0]>0xff9f)
	return -1;
    char dh = voiced_table[kata_hira][src[0]-0xff61];
    if(combine && (dh & D) && src[1]==U16HAN_VOICEDSOUNDMARK)
	return 1;
    if(combine && (dh & H) && src[1]==U16HAN_SEMIVOICEDSOUNDMARK)
	return 2;
    return 0;
    #undef D
    #undef H
}

/*
  u16Ⱦѥʢu16Ѥ/ʡcombinetrueʤϹ롣
  srclenꤷȤϥ̥ʸ̵뤹롣
  u16ʤΤ''롣ejѴƦʸˤʤΤդ뤳ȡ''ejˤϤʤ
  kata_hira:Ҥ餬ʤλ1,ʤλ0
 */
uint16_t* U16HanToZen(uint16_t* dst,int* dstlen,const uint16_t* src,int srclen,int kata_hira,bool combine)
{
    if(src==NULL || srclen==0)
	return NULL;
    
    const uint16_t hk2zk[]={ //u-ff61 
	0x3002/**/,	0x300c/**/,	0x300d/**/,	0x3001/**/,	0x30fb/**/,
	0x30f2/**/,	0x30a1/**/,	0x30a3/**/,	0x30a5/**/,	0x30a7/**/,	0x30a9/**/,
	0x30e3/**/,	0x30e5/**/,	0x30e7/**/,	0x30c3/**/,	0x30fc/**/, 	0x30a2/**/,
	0x30a4/**/,	0x30a6/**/,	0x30a8/**/,	0x30aa/**/,	0x30ab/**/,	0x30ad/**/,
	0x30af/**/,	0x30b1/**/,	0x30b3/**/,	0x30b5/**/, 	0x30b7/**/,	0x30b9/**/,
	0x30bb/**/,	0x30bd/**/,	0x30bf/**/,	0x30c1/**/,	0x30c4/**/,	0x30c6/**/,
	0x30c8/**/,	0x30ca/**/,	0x30cb/**/, 	0x30cc/**/,	0x30cd/**/,	0x30ce/**/,
	0x30cf/**/,	0x30d2/**/,	0x30d5/**/,	0x30d8/**/, 	0x30db/**/,	0x30de/**/,
	0x30df/**/,	0x30e0/**/,	0x30e1/**/,	0x30e2/**/,	0x30e4/**/,	0x30e6/**/,
	0x30e8/**/,	0x30e9/**/, 	0x30ea/**/,	0x30eb/**/,	0x30ec/**/,	0x30ed/**/,
	0x30ef/**/,	0x30f3/**/, 	0x309b/**/, 	0x309c/**/ /*u-ff9f*/
    };
    const uint16_t hk2zh[]={ //u-ff61 Ҥ餬   ĹȾϥʤƱ
	0x3002/**/,	0x300c/**/,	0x300d/**/,	0x3001/**/,	0x30fb/**/,
	0x3092/**/,	0x3041/**/,	0x3043/**/,	0x3045/**/,	0x3047/**/,	0x3049/**/,
	0x3083/**/,	0x3085/**/,	0x3087/**/,	0x3063/**/,	0x30fc/**/, 	0x3042/**/,
	0x3044/**/,	0x3046/**/,	0x3048/**/,	0x304a/**/,	0x304b/**/,	0x304d/**/,
	0x304f/**/,	0x3051/**/,	0x3053/**/,	0x3055/**/, 	0x3057/**/,	0x3059/**/,
	0x305b/**/,	0x305d/**/,	0x305f/**/,	0x3061/**/,	0x3064/**/,	0x3066/**/,
	0x3068/**/,	0x306a/**/,	0x306b/**/, 	0x306c/**/,	0x306d/**/,	0x306e/**/,
	0x306f/**/,	0x3072/**/,	0x3075/**/,	0x3078/**/, 	0x307b/**/,	0x307e/**/,
	0x307f/**/,	0x3080/**/,	0x3081/**/,	0x3082/**/,	0x3084/**/,	0x3086/**/,
	0x3088/**/,	0x3089/**/, 	0x308a/**/,	0x308b/**/,	0x308c/**/,	0x308d/**/,
	0x308f/**/,	0x3093/**/, 	0x309b/**/, 	0x309c/**/ /*u-ff9f*/
    };
    const uint16_t* hk2z[]={hk2zk,hk2zh};
    const uint16_t daku[]={ //u-ff61
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0x30fa/**/,	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/, 	0/**/,
	0/**/,		0x30f4/**/,	0/**/,		0/**/,		0x30ac/**/,	0x30ae/**/,
	0x30b0/**/,	0x30b2/**/,	0x30b4/**/,	0x30b6/**/, 	0x30b8/**/,	0x30ba/**/,
	0x30bc/**/,	0x30be/**/,	0x30c0/**/,	0x30c2/**/,	0x30c5/**/,	0x30c7/**/,
	0x30c9/**/,	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0x30d0/**/,	0x30d3/**/,	0x30d6/**/,	0x30d9/**/, 	0x30dc/**/,	0/**/,
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0/**/,		0/**/, 	0/**/,		0/**/,		0/**/,		0/**/,
	0x30f7/**/,	0/**/, 	0/**/, 	0/**/ /*u-ff9f*/
    };
    const uint16_t handaku[]={ //u-ff61
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/, 	0/**/,
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0/**/,		0/**/,		0/**/,		0/**/, 	0/**/,		0/**/,
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0x30d1/**/,	0x30d4/**/,	0x30d7/**/,	0x30da/**/, 	0x30dd/**/,	0/**/,
	0/**/,		0/**/,		0/**/,		0/**/,		0/**/,		0/**/,
	0/**/,		0/**/, 	0/**/,		0/**/,		0/**/,		0/**/,
	0/**/,		0/**/, 	0/**/, 	0/**/ /*u-ff9f*/
    };

    if(dst == NULL)
	dst = calloc((srclen>0 ? srclen:WcLen(src))+1,2);
    uint16_t* dst0 = dst;

    const uint16_t begin=0xff61;
    uint16_t kata_to_hira = 0x60*kata_hira; //ʢҤ ؤΥեå
    uint16_t let;
    while(let=*src, srclen>0 || (srclen<0 && let!=0)){
	switch(U16CombineHan(src++,kata_hira,combine)){
	case 0:
	    if(let == 0x20) //space
		*dst = 0x3000;
	    else if(let < 0x80) //ascii !!!ȥʸϹθƤʤ
		*dst = let-0x20+0xff00;
	    else //Ⱦѥ
		*dst = hk2z[kata_hira][let-begin];
	    break;
	case 1:
	    *dst = daku[let-begin] - kata_to_hira;
	    src++;
	    break;
	case 2:
	    *dst = handaku[let-begin] - kata_to_hira;
	    src++;
	    break;
	default:
	    *dst = let;
	}
	++dst;
	--srclen;
    }
    *dst = 0; //ûʤǽΤǥ̥ʸƤ
    if(dstlen != NULL)
	*dstlen = dst-dst0;
    return dst0;
}

/*
  u16ѤҤ餬/ʢȾѥ
  srclen<0ΤȤdstlen˥̥ʸޤࡣsrclenꤷȤϤϰϤǽʸˤʤ롣
*/
uint16_t* U16ZenToHan(uint16_t* dst,int* dstlen,const uint16_t* src,int srclen)
{
    if(src == NULL)
	return NULL;
    
    const int errlet=0xfffd;
    const uint32_t zenkana_to_han[]={ //Ⱦо壱ӥåȤ˥ɤ롣
	0xff67/* 0x3041*/,
	0xff71/**/,	0xff68/**/,	0xff72/**/,	0xff69/**/,	0xff73/**/,
 	0xff6a/**/,	0xff74/**/,	0xff6b/**/,	0xff75/**/, 	0xff76/**/,
	0xff76|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff77/**/,
 	0xff77|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff78/**/,
	0xff78|(U16HAN_VOICEDSOUNDMARK<<16)/**/, 	0xff79/**/,
 	0xff79|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff7a/**/,
 	0xff7a|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff7b/**/,
 	0xff7b|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff7c/**/,
 	0xff7c|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff7d/**/,
 	0xff7d|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff7e/**/,
 	0xff7e|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff7f/**/,
 	0xff7f|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff80/**/,
 	0xff80|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff81/**/,
 	0xff81|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff6f/**/,
 	0xff82/**/,	0xff82|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff83/**/,
 	0xff83|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff84/**/,
 	0xff84|(U16HAN_VOICEDSOUNDMARK<<16)/**/,	0xff85/**/,	0xff86/**/,
 	0xff87/**/,	0xff88/**/,	0xff89/**/,	0xff8a/**/,
 	0xff8a|(U16HAN_VOICEDSOUNDMARK<<16)/**/,
	0xff8a|(U16HAN_SEMIVOICEDSOUNDMARK<<16)/**/,	0xff8b/**/,
 	0xff8b|(U16HAN_VOICEDSOUNDMARK<<16)/**/,
 	0xff8b|(U16HAN_SEMIVOICEDSOUNDMARK<<16)/**/,	0xff8c/**/,
 	0xff8c|(U16HAN_VOICEDSOUNDMARK<<16)/**/,
 	0xff8c|(U16HAN_SEMIVOICEDSOUNDMARK<<16)/**/,	0xff8d/**/,
 	0xff8d|(U16HAN_VOICEDSOUNDMARK<<16)/**/,
 	0xff8d|(U16HAN_SEMIVOICEDSOUNDMARK<<16)/**/,	0xff8e/**/,
 	0xff8e|(U16HAN_VOICEDSOUNDMARK<<16)/**/,
 	0xff8e|(U16HAN_SEMIVOICEDSOUNDMARK<<16)/**/,	0xff8f/**/,	0xff90/**/,
	0xff91/**/,	0xff92/**/,	0xff93/**/,	0xff6c/**/,	0xff94/**/,
 	0xff6d/**/,	0xff95/**/,	0xff6e/**/,	0xff96/**/,	0xff97/**/,
 	0xff98/**/,	0xff99/**/,	0xff9a/**/,	0xff9b/**/, 	errlet/**/,
 	0xff9c/**/, 	errlet/*<?>*/, 		errlet/*<?>*/,
	0xff66/**/, 	0xff9d/**/,
 	0xff73|(U16HAN_VOICEDSOUNDMARK<<16)/* 0x3094*/,
    };

    if(dst == NULL)
	dst = calloc((srclen>0 ? srclen:WcLen(src))*2+1,2); //ΤᣲܳݤƤ
    uint16_t* dst0 = dst;
    uint16_t let;
    while(let=*(src++), srclen>0 || (srclen<0 && let!=0)){
	switch(let){
	case 0x30a1 ... 0x30f4:/*..*/
	    let -= 0x60;
	case 0x3041 ... 0x3094:/*..*/{
	    uint32_t z = zenkana_to_han[let-0x3041];
	    *dst = (z & 0xffff);
	    if((z & 0xffff0000) != 0)
		*(++dst) = (z>>16);
	    break;
	}
	case 0x3001:/**/
	    *dst = 0xff64;
	    break;
	case 0x3002:/**/
	    *dst = 0xff61;
	    break;
	case 0x300c:/**/
	    *dst = 0xff62;
	    break;
	case 0x300d:/**/
	    *dst = 0xff63;
	    break;
	case 0x30fb: /**/
	    *dst = 0xff65;
	    break;
	case 0x309b:/*  */
	    *dst = U16HAN_VOICEDSOUNDMARK;
	    break;
	case 0x309c:/*   */
	    *dst = U16HAN_SEMIVOICEDSOUNDMARK;
	    break;
	case 0x30fc:/**/
	    *dst = 0xff70;
	    break;
	case 0x3000:/**/
	    *dst = 0x20;
	    break;
	case 0xff01 ... 0xff5e:/* .. */
	    *dst = (let&0xff)+0x20;
	    break;
	default:
	    *dst = errlet/*<?>*/;
	}
	--srclen;
	++dst;
    }
    if(srclen < 0)
	*(dst++) = 0;
    if(dstlen != NULL)
	*dstlen = dst-dst0;
    return dst0;
}

//utf8 --> euc-jp
char* U8ToEj(char* dst,const char* src)
{
    if(src == NULL)
	return NULL;

    size_t ileft=strlen(src),oleft=ileft+1;
    char *dst0,*src0,*srcp;

    src0 = srcp = strdup(src);
    if(dst == NULL)
	dst = malloc(oleft+1);
    dst0 = dst;
    while(!conv(U8E_08,&srcp,&ileft,&dst,&oleft)){
	*(dst++) = (char)(TOFU_E & 0xff);
	*(dst++) = (char)(TOFU_E >> 8);
	oleft -= 2;
	char* srcerr = srcp;
	srcp = ForwardU8(srcp,1);
	ileft -= srcp-srcerr;
    }
    *dst = 0;
    free(src0);
    return dst0;
}

//utf8ʸƬnʸư롣
char* ForwardU8(const char* str,int n)
{
    while(--n >= 0 && *str!=0){
	unsigned char ch = *(str++);
	if((ch & 0x80)!=0){
	    while(((ch <<= 1) & 0x80) != 0)
		++str;
	}
    }
    return (char*)str;
}

uint16_t* U8ToU16(uint16_t* out,const char* in)
{
    if(in == NULL)
	return NULL;
    
    size_t ileft = strlen(in);
    size_t oleft = ileft*2;
    if(out == NULL)
	out = calloc(ileft+1,2);
    uint16_t* out0 = out;
    char* inwk = strdup(in);
    char* inwk0 = inwk;
    conv(U8U16,&inwk,&ileft,(char**)&out,&oleft);
    *out = 0;
    free(inwk0);
    return out0;
}

char* StrDel(char* str,int pos,int len)
{
    memmove(str+pos,str+pos+len,strlen(str+pos+len)+1);
    return str;
}

//(C) 2008 thomas
